@@ -1,7 +1,14 @@
+import pickle
+import random
+from time import time
+from typing import Union
+
 import numpy as np
 import torch
-from torch import nn
+from torch import nn, optim
 from torch.nn import functional as F
+from torch.nn.functional import mse_loss, l1_loss, binary_cross_entropy, cross_entropy
+from torch.optim import Optimizer
 
 
 class NBeatsNet(nn.Module):
@@ -10,12 +17,12 @@ class NBeatsNet(nn.Module):
     GENERIC_BLOCK = 'generic'
 
     def __init__(self,
-                 device,
+                 device=torch.device('cpu'),
                  stack_types=(TREND_BLOCK, SEASONALITY_BLOCK),
                  nb_blocks_per_stack=3,
                  forecast_length=5,
                  backcast_length=10,
-                 thetas_dims=(4, 8),
+                 thetas_dim=(4, 8),
                  share_weights_in_stack=False,
                  hidden_layer_units=256,
                  nb_harmonics=None):
@@ -28,7 +35,7 @@ class NBeatsNet(nn.Module):
         self.nb_harmonics = nb_harmonics
         self.stack_types = stack_types
         self.stacks = []
-        self.thetas_dim = thetas_dims
+        self.thetas_dim = thetas_dim
         self.parameters = []
         self.device = device
         print('| N-Beats')
@@ -36,6 +43,8 @@ class NBeatsNet(nn.Module):
             self.stacks.append(self.create_stack(stack_id))
         self.parameters = nn.ParameterList(self.parameters)
         self.to(self.device)
+        self._loss = None
+        self._opt = None
 
     def create_stack(self, stack_id):
         stack_type = self.stack_types[stack_id]
@@ -53,6 +62,13 @@ class NBeatsNet(nn.Module):
             blocks.append(block)
         return blocks
 
+    def save(self, filename: str):
+        torch.save(self, filename)
+
+    @staticmethod
+    def load(f, map_location=None, pickle_module=pickle, **pickle_load_args):
+        return torch.load(f, map_location, pickle_module, **pickle_load_args)
+
     @staticmethod
     def select_block(block_type):
         if block_type == NBeatsNet.SEASONALITY_BLOCK:
@@ -62,7 +78,91 @@ class NBeatsNet(nn.Module):
         else:
             return GenericBlock
 
+    def compile(self, loss: str, optimizer: Union[str, Optimizer]):
+        if loss == 'mae':
+            loss_ = l1_loss
+        elif loss == 'mse':
+            loss_ = mse_loss
+        elif loss == 'cross_entropy':
+            loss_ = cross_entropy
+        elif loss == 'binary_crossentropy':
+            loss_ = binary_cross_entropy
+        else:
+            raise ValueError(f'Unknown loss name: {loss}.')
+        # noinspection PyArgumentList
+        if isinstance(optimizer, str):
+            if optimizer == 'adam':
+                opt_ = optim.Adam
+            elif optimizer == 'sgd':
+                opt_ = optim.SGD
+            elif optimizer == 'rmsprop':
+                opt_ = optim.RMSprop
+            else:
+                raise ValueError(f'Unknown opt name: {optimizer}.')
+            opt_ = opt_(lr=1e-4, params=self.parameters())
+        else:
+            opt_ = optimizer
+        self._opt = opt_
+        self._loss = loss_
+
+    def fit(self, x_train, y_train, validation_data=None, epochs=10, batch_size=32):
+
+        def split(arr, size):
+            arrays = []
+            while len(arr) > size:
+                slice_ = arr[:size]
+                arrays.append(slice_)
+                arr = arr[size:]
+            arrays.append(arr)
+            return arrays
+
+        for epoch in range(epochs):
+            x_train_list = split(x_train, batch_size)
+            y_train_list = split(y_train, batch_size)
+            assert len(x_train_list) == len(y_train_list)
+            shuffled_indices = list(range(len(x_train_list)))
+            random.shuffle(shuffled_indices)
+            self.train()
+            train_loss = []
+            timer = time()
+            for batch_id in shuffled_indices:
+                batch_x, batch_y = x_train_list[batch_id], y_train_list[batch_id]
+                self._opt.zero_grad()
+                _, forecast = self(torch.tensor(batch_x, dtype=torch.float).to(self.device))
+                loss = self._loss(forecast, squeeze_last_dim(torch.tensor(batch_y, dtype=torch.float).to(self.device)))
+                train_loss.append(loss.item())
+                loss.backward()
+                self._opt.step()
+            elapsed_time = time() - timer
+            train_loss = np.mean(train_loss)
+
+            test_loss = '[undefined]'
+            if validation_data is not None:
+                x_test, y_test = validation_data
+                self.eval()
+                _, forecast = self(torch.tensor(x_test, dtype=torch.float).to(self.device))
+                test_loss = self._loss(forecast, squeeze_last_dim(torch.tensor(y_test, dtype=torch.float))).item()
+
+            num_samples = len(x_train_list)
+            time_per_step = int(elapsed_time / num_samples * 1000)
+            print(f'Epoch {str(epoch + 1).zfill(len(str(epochs)))}/{epochs}')
+            print(f'{num_samples}/{num_samples} [==============================] - '
+                  f'{int(elapsed_time)}s {time_per_step}ms/step - '
+                  f'loss: {train_loss:.4f} - val_loss: {test_loss:.4f}')
+
+    def predict(self, x, return_backcast=False):
+        self.eval()
+        b, f = self(torch.tensor(x, dtype=torch.float).to(self.device))
+        b, f = b.detach().numpy(), f.detach().numpy()
+        if len(x.shape) == 3:
+            b = np.expand_dims(b, axis=-1)
+            f = np.expand_dims(f, axis=-1)
+        if return_backcast:
+            return b
+        return f
+
     def forward(self, backcast):
+        backcast = squeeze_last_dim(backcast)
         forecast = torch.zeros(size=(backcast.size()[0], self.forecast_length,))  # maybe batch size here.
         for stack_id in range(len(self.stacks)):
             for block_id in range(len(self.stacks[stack_id])):
@@ -72,6 +172,12 @@ class NBeatsNet(nn.Module):
         return backcast, forecast
 
 
+def squeeze_last_dim(tensor):
+    if len(tensor.shape) == 3 and tensor.shape[-1] == 1:  # (128, 10, 1) => (128, 10).
+        return tensor[..., 0]
+    return tensor
+
+
 def seasonality_model(thetas, t, device):
     p = thetas.size()[-1]
     assert p <= thetas.shape[1], 'thetas_dim is too big.'
@@ -89,10 +195,10 @@ def trend_model(thetas, t, device):
     return thetas.mm(T.to(device))
 
 
-def linspace(backcast_length, forecast_length):
-    lin_space = np.linspace(-backcast_length, forecast_length, backcast_length + forecast_length)
-    b_ls = lin_space[:backcast_length]
-    f_ls = lin_space[backcast_length:]
+def linear_space(backcast_length, forecast_length):
+    ls = np.arange(-backcast_length, forecast_length, 1) / forecast_length
+    b_ls = ls[:backcast_length]
+    f_ls = ls[backcast_length:]
     return b_ls, f_ls
 
 
@@ -111,7 +217,7 @@ class Block(nn.Module):
         self.fc3 = nn.Linear(units, units)
         self.fc4 = nn.Linear(units, units)
         self.device = device
-        self.backcast_linspace, self.forecast_linspace = linspace(backcast_length, forecast_length)
+        self.backcast_linspace, self.forecast_linspace = linear_space(backcast_length, forecast_length)
         if share_thetas:
             self.theta_f_fc = self.theta_b_fc = nn.Linear(units, thetas_dim, bias=False)
         else:
@@ -119,6 +225,7 @@ class Block(nn.Module):
             self.theta_f_fc = nn.Linear(units, thetas_dim, bias=False)
 
     def forward(self, x):
+        x = squeeze_last_dim(x)
         x = F.relu(self.fc1(x.to(self.device)))
         x = F.relu(self.fc2(x))
         x = F.relu(self.fc3(x))
